---
jupytext:
  formats: md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.13
    jupytext_version: 1.16.4
kernelspec:
  display_name: travo-kernel-2
  language: python
  name: travo-kernel-2
---

## Exercice N : jeu de cartes
On souhaite modéliser un jeu de cartes et estimer numériquement la probabilité de piocher au moins une figure (valet ou "J", reine ou "Q", roi ou "K", as ou "A") dans une main initiale de 5 cartes, tirée au hasard et sans remplacement.

Les valeurs possibles pour les cartes qui ne sont pas des figures vont de 2 à 10.

```{code-cell} ipython3
---
editable: false
slideshow:
  slide_type: ''
---
import random, numpy as np, matplotlib.pyplot as plt
```

```{code-cell} ipython3
---
editable: false
slideshow:
  slide_type: ''
---
liste_figures = ["J", "Q", "K", "A"]
liste_valeurs = list(range(2, 11)) + liste_figures
N_valeurs = len(liste_valeurs)
print("Voici les {} valeurs possibles : {}".format(N_valeurs, liste_valeurs))
```

+++ {"editable": true, "slideshow": {"slide_type": ""}}

1. Écrivez une classe `Carte` qui modélise une carte, avec les propriétés suivantes :
   * Un attribut `valeur` défini à l'initialisation de l'instance (ou *instanciation*) ;
   * Une méthode `figure` qui renvoie `True` si `self.valeur` fait partie des valeurs possibles pour les figures (servez-vous de la liste `liste_figures` définie plus haut) ;
   * Une méthode `__repr__` qui renvoie la valeur de la carte **convertie en chaîne de caractères**.

```{code-cell} ipython3
---
editable: true
slideshow:
  slide_type: ''
---
### BEGIN SOLUTION
class Carte():
    def __init__(self, valeur):
        self.valeur = valeur
    def figure(self):
        return self.valeur in liste_figures
    def __repr__(self):
        return str(self.valeur)
### END SOLUTION
```

+++ {"editable": true, "slideshow": {"slide_type": ""}}

2. Écrivez une classe `Main` qui modélise une main initiale avec un nombre $k$ de cartes piochées, avec les propriétés suivantes :
   * Un attribut `k`, défini à l'instanciation, qui représente le nombre de cartes piochées ;
   * Un attribut `cartes` (une liste), défini à l'instanciation, obtenu à partir du tirage aléatoire et sans remplacement de $k$ cartes (utiliser `random.sample` avec les bons arguments), chaque élément de cette liste doit être une instance de la classe `Carte` ;
   * Une méthode `contient_figure` qui renvoie `True` si une figure a été piochée parmi les cartes en main (vous pouvez par exemple utiliser `np.any` avec les bons arguments) ;
   * Une méthode `__repr__` qui renvoie une chaîne de caractères indiquant le nombre de cartes en main et la liste des cartes piochées avec leur valeur.

Lors de l'instanciation (méthode `__init__`) on s'assurera que le nombre de cartes tirées $k$ est inférieur ou égal au nombre de cartes qu'il est possible de tirer (`N_valeurs` défini plus haut).

```{code-cell} ipython3
---
editable: true
slideshow:
  slide_type: ''
---
### BEGIN SOLUTION
class Main():
    def __init__(self, k=5):
        assert k <= N_valeurs
        self.k = k
        choix = random.sample(liste_valeurs, k=k)
        self.cartes = [Carte(carte) for carte in choix]
    def contient_figure(self):
        return np.any([carte.figure() for carte in self.cartes])
    def __repr__(self):
        return "Main de {} cartes: {}".format(self.k, self.cartes)
### END SOLUTION
```

3. Testez votre classe `Main` en tirant $k=5$ cartes aléatoires :
   * Affichez l'objet ainsi créé (avec `print`) et vérifiez que ce qui s'affiche correspond bien à ce que vous avez précédemment programmé dans les méthodes `__repr__` ;
   * Affichez également le résultat de la méthode `contient_figure` et vérifiez que la réponse donnée a un sens (vous pouvez ré-exécuter la cellule plusieurs fois jusqu'à tomber sur une main sans figures et vérifier que la méthode renvoie bien `False`).

```{code-cell} ipython3
---
editable: true
slideshow:
  slide_type: ''
---
### BEGIN SOLUTION
m = Main(5)
print(m, m.contient_figure())
### END SOLUTION
```

4. Effectuez $M=10 000$ tirages de $k=5$ cartes chacun en utilisant votre classe `Main`. Quelle est la probabilité de piocher au moins 1 figure pour une main initiale de 5 cartes ?

```{code-cell} ipython3
### BEGIN SOLUTION
res = 0
k = 5
M = 10000
for i in range(M):
    m = Main(k)
    res += int(m.contient_figure())
print("Probabilité de piocher au moins 1 figure sur {} cartes dans la main initiale : {:.4f}".format(k, res/M))
### END SOLUTION
```

5. Faites une boucle sur la valeur de $M$ (en prenant les valeurs 1, 2, 5, 10, 100, 1 000, 2 000, 5 000 et 10 000) et tracez, en échelle semi-logarithmique (`plt.semilogx`), la probabilité de piocher au moins 1 figure pour une main initiale de $k=5$ cartes en fonction du nombre de tirages, $M$. **Veillez à soigner le graphe (titres des axes, grille, etc.).**

```{code-cell} ipython3
---
editable: true
slideshow:
  slide_type: ''
---
### BEGIN SOLUTION
p_num = []
M = np.array([1, 2, 5, 10, 100, 1000, 2000, 5000, 10000])
for i in M:
    res = 0
    for j in range(i):
        m = Main(5)
        res += int(m.contient_figure())
    p_num.append(res/i)

p_num = np.array(p_num)

plt.semilogx(M, p_num)
plt.xlabel("$M$")
plt.ylabel(r"$p_\mathrm{num}$")
plt.grid(True)
### END SOLUTION
```

6. **Défi (bonus) :**

On souhaite comparer nos résultats avec la prédiction théorique (distribution hypergéométrique) pour piocher au moins 1 figure en tirant $k=5$ cartes parmi $N=13$, sachant qu'il y a $F=4$ figures :

$$
p_\mathrm{th} = P(s\geq 1;k,F,N) = \displaystyle\sum_{s=1}^F \dfrac{\binom{F}{s} \binom{N-F}{k-s}}{\binom{N}{k}}
$$

où $\binom{j}{i}$ représente la loi binomiale (nombre de combinaisons de $i$ éléments parmi $j$ sans remplacement).

 * Représentez, en échelle logarithmique (`plt.loglog`), la valeur absolue de l'erreur relative $\epsilon$ entre le résultat numérique $p_\mathrm{num}$ et le résultat théorique $p_\mathrm{th}$ en fonction du nombre de lancers $M$ pour les valeurs de la question 5 :

$$
\epsilon = \left|\dfrac{p_\mathrm{num} - p_\mathrm{th}}{p_\mathrm{th}}\right|
$$

```{code-cell} ipython3
### BEGIN SOLUTION
from scipy.special import binom
F = 4
N = 13
k = 5
p_th = np.sum([binom(F, s) * binom(N-F, k-s) / binom(N, k) for s in range(1, F+1)])
epsilon = np.abs((p_num - p_th) / p_th)

plt.loglog(M, epsilon)
plt.xlabel("M")
plt.ylabel(r"$\epsilon$")
plt.grid(True)
### END SOLUTION
```

```{code-cell} ipython3
---
editable: true
slideshow:
  slide_type: ''
---

```
